from fabric import api


def local_run_wrapper(*args, **kwargs):
    """
    Wrapper around fabric's `local` command with the capture parameter always
    enabled.
    """
    kwargs['capture'] = True

    return api.local(*args, **kwargs)



"""
This module provides a deployment structure similar to what Capistrano does. By
default, the project layout looks like that::

    current -> releases/20140830180015_1.2.3  -- Symlink to the current release
    releases/
        20140830151210_1.2.2/
        20140831180015_1.2.3/
    repository.git/                           -- Git repository containing the project
    shared/                                   -- Shared files not specific to a release
        config.yml
        media/
    VERSION                                   -- The version of the current release

When you create a new release, the code from the `repository.git` directory
is archived in a new directory named after the current date and the version
you're deploying. After that, the symlink `current` is switched to the new
release.

The following variables must be defined in Fabric's env for this module to
work:

`releases_root`
    Path to the releases/ directory

`repository_root`
    Path to the repository.git/ directory

`shared_root`
    Path to the shared/ directory

`shared_files`
    Dictionary of shared files {target: link_name}

The following variable should be defined if you want to use the various methods
of this module without having to pass it around :

`release_name`
    Name of the release
"""

from contextlib import nested
import logging
import os

from fabric.api import cd, env, run
from fabric.context_managers import quiet

from . import signals
from .file import ls


FAILED_RELEASE_SUFFIX = '_failed'


logger = logging.getLogger(__name__)


def determine_release_name(release_name):
    """
    Try to get ``release_name`` from :py:attr:`fabric.api.env.release_name`.
    """
    if release_name is None:
        if hasattr(env, 'release_name'):
            release_name = env.release_name
        else:
            raise AttributeError

    return release_name


def get_release_path(release_name=None):
    """
    Return the absolute path to the directory of the given release.

    If ``release_name is`` not given, try to get it from :py:attr:`fabric.api.env.release_name`.

    Arguments:
        release_name -- The name of the release (usually a date like YmdHMS)
    """
    release_name = determine_release_name(release_name)
    return os.path.join(env.releases_root, release_name)


@signals.register
def create_release(tag, release_name=None):
    """
    Create the directory for a new release and extract the contents from the
    git repository at the given tag and put them in this directory.

    Arguments:
        release_name -- The name of the release (usually a date like YmdHMS)
        tag -- The tag to install in this release
    """
    release_path = get_release_path(release_name)

    run("mkdir %s" % release_path)
    tmpfile = run("mktemp")
    run("git archive --output={tmpfile} --remote={remote} {version}"
        " && tar -x -C {tmpfile} {release_path}".format(
            remote=env.repository_root,
            version=tag,
            release_path=release_path,
            tmpfile=tmpfile))
    run("rm -f {tmpfile}".format(tmpfile=tmpfile))


@signals.register
def link_shared_files(release_name=None):
    """
    Create or update links to shared files defined in the ``shared_files`` env
    variable.

    If ``release_name is`` not given, try to get it from :py:attr:`fabric.api.env.release_name`.

    Arguments:
        release_name -- The name of the release (usually a date like YmdHMS)
    """
    release_path = get_release_path(release_name)

    for target, link_path in env.shared_files.iteritems():
        target_abspath = os.path.join(env.shared_root, target)
        link_path = os.path.join(release_path, link_path)

        run("ln -s {target} {link_path}".format(
            target=target_abspath, link_path=link_path))


@signals.register
def activate_release(release_name=None):
    """
    Activate the given release by making the ``current`` symlink point to it.

    If ``release_name is`` not given, try to get it from :py:attr:`fabric.api.env.release_name`.

    Arguments:
        release_name -- The name of the release (usually a date like YmdHMS)
    """
    logger.debug("""

              ~ RELEASE THE KRAKEN!!! ~
                        ___
                     .-'   `'.
                    /         \\
                    |         ;
                    |         |           ___.--,
           _.._     |0) ~ (0) |    _.---'`__.-( (_.
    __.--'`_.. '.__.\\    '--. \\_.-' ,.--'`     `""`
   ( ,.--'`   ',__ /./;   ;, '.__.'`    __
   _`) )  .---.__.' / |   |\\   \\__..--""  \"\""--.,_
  `---' .'.''-._.-'`_./  /\\ '.  \\ _.-~~~````~~~-._`-.__.'
        | |  .' _.-' |  |  \\  \\  '.               `~---`
         \\ \\/ .'     \\  \\   '. '-._)
          \\/ /        \\  \\    `=.__`~-.
     jgs  / /\\         `) )    / / `"".`\\
    , _.-'.'\\ \\        / /    ( (     / /
     `--~`   ) )    .-'.'      '.'.  | (
            (/`    ( (`          ) )  '-;
             `      '-;         (-'
    """)

    with cd(env.project_root):
        run("ln -s {target} new_current".format(
            target=get_release_path(release_name)))
        run("mv -Tf new_current current")


@signals.register
def clean_old_releases(keep=5):
    """
    Remove the old release directories from the releases directory, keeping x
    releases defined by the ``keep`` parameter.

    Arguments:
        keep -- The number of releases to keep
    """
    releases = get_releases()
    status = True

    for release in releases[:-keep]:
        status = run("rm -rf %s" % get_release_path(release), warn_only=True)
        if status.return_code != 0:
            status = False
    return status


def invalidate_last_release():
    """
    Invalidate the last made release so that it won't be a target for a rollback.
    """
    releases = get_releases()
    last_release = releases[-1]
    run("mv {release} {release}{suffix}".format(
        release=get_release_path(last_release),
        suffix=FAILED_RELEASE_SUFFIX))


def get_releases():
    """
    Return the list of releases on the server, sorted by oldest to newest.
    """
    return filter(
        lambda f: not f.endswith(FAILED_RELEASE_SUFFIX),
        sorted(ls(os.path.join(env.releases_root)))
    )


def get_currently_installed_version():
    """
    Return the currently installed version (tag) by reading the contents of the
    VERSION file, or None if the VERSION file could not be read.
    """
    with nested(cd(env.project_root), quiet()):
        version = run("cat VERSION")

    return version if version.succeeded else None


@signals.register
def update_version_file(version):
    """
    Update the VERSION file with the given version.
    """
    with nested(cd(env.project_root), quiet()):
        run("echo %s > VERSION" % version)



from contextlib import nested

from fabric import api
from fabric.context_managers import quiet


def push_tag(tag, remote='origin'):
    """
    Pushes the given tag to the given remote.
    """
    api.local('git push {remote} {tag}'.format(
        remote=remote,
        tag=tag)
    )


def update_remote_repository_root(tag):
    """
    Fetches the latest git objects on the remote, checks out the given tag and
    updates the submodules if necessary.

    Requires the `repository_root` environment variable to be set.
    """
    with nested(api.cd(api.env.repository_root), api.hide('commands')):
        api.run('git fetch -t -p')
        api.run('git checkout {tag}'.format(tag=tag))
        api.run('git submodule sync')
        api.run('git submodule update --init')


def get_latest_tag(commit='HEAD', run_locally=True):
    """
    Return the latest reachable tag from the given commit.

    Arguments:
    commit -- The name of the commit to use (tag, hash, etc) (default HEAD)
    run_locally -- Whether to get the latest local or remote tag (default True)

    """
    git_command = 'git describe --tag --abbrev=0 {commit}'.format(
        commit=commit if commit is not None else ''
    )

    with nested(api.hide('commands'), quiet()):
        if run_locally:
            tag = api.local(git_command, capture=True)
        else:
            with api.cd(api.env.repository_root):
                tag = api.run(git_command)

    # For strange reasons the above call to run returns git's stderr in
    # case of failure (eg. if there are no tags yet) even with combine_stderr
    # set to False. This is a temporary hack to avoid returning git's error
    # message as the tag name
    if tag.return_code != 0:
        tag = ''

    return tag


def get_latest_commit(run_locally=True):
    """
    Return the commit identified by the current HEAD.

    Arguments:
        run_locally -- Whether to get the latest local or remote HEAD (default
        True)

    """
    git_command = 'git rev-parse HEAD'

    with nested(api.hide('commands'), quiet()):
        if run_locally:
            commit = api.local(git_command, capture=True)
        else:
            with api.cd(api.env.repository_root):
                commit = api.run(git_command)

    return commit


def get_commit_messages(first_commit, last_commit):
    """
    Returns all commit messages between first_commit and last_commit in an
    abbreviated form.
    """
    with api.hide('commands'):
        changes = api.local(
            'git log --reverse --oneline {first_commit}..{last_commit}'
            .format(
                first_commit=first_commit,
                last_commit=last_commit
            ), capture=True
        )

    return changes



from contextlib import nested

from fabric import api
from fabric.context_managers import quiet


def ls(path):
    """
    Return the list of the files in the given directory, omitting . and ...

    Arguments:
        path -- The path of the directory to get the files from
    """
    with nested(api.cd(path), quiet()):
        files = api.run('for i in *; do echo $i; done')
        files_list = files.replace('\r', '').split('\n')

    return files_list


def file_exists(path):
    """
    Checks if the given path exists on the host and returns True if that's the
    case, False otherwise.
    """
    with quiet():
        exists = api.run('test -e {path}'.format(path=path)).succeeded

    return exists



from functools import wraps

from fabric import api


def multisite(func):
    """
    Mark a task as being multisite.

    Multisite tasks allow you to define a global configuration for multiple
    sites and specify the site to use when you run your command. All the
    dictionary keys of the given site will then be made available in the `env`
    variable. The structure is
    `env.sites[site_name][env_name][configuration_key] = configuration_value`,
    where `env_name` is the name of the decorated function. Also the decorated
    function must take a parameter `site`.

    Here's an example with sites A and B, both having prod and staging
    environments::

        env.sites = {
            'site_a': {
                'prod': {
                    'project_root': '/var/www/site_a/prod/'
                },
                'staging': {
                    'project_root': '/var/www/site_a/staging/'
                }
            },
            'site_b': {
                'prod': {
                    'project_root': '/var/www/site_b/prod/'
                },
                'staging': {
                    'project_root': '/var/www/site_b/staging/'
                }
            }
        }

        @task
        @multisite
        def prod(site):
            # You could also define hosts in the `env.sites` dictionary if
            # they're different from one site to another
            # If the site is set to `site_a`, the decorator will take the
            # contents from env.sites['site_a']['prod'] and make it available
            # in the `env` variable
            env.hosts = ['www.my_host.com']

        @task
        @multisite
        def staging(site):
            # You could also define hosts in the `env.sites` dictionary if
            # they're different from one site to another
            env.hosts = ['staging.my_host.com']
    """

    if 'sites' not in api.env:
        raise RuntimeError("To use the multisite decorator you need to set the"
                           " `env.sites` variable.")

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args:
            site = args[0]
        elif func.__defaults__:
            site = func.__defaults__[0]
        else:
            site = None

        selected_environment = func.__name__

        if site is not None:
            if site not in api.env.sites:
                raise Exception(
                    "Site {site} is not part of the possible sites ({sites})"
                    .format(site=site, sites=api.env.sites.keys())
                )

            if selected_environment not in api.env.sites[site]:
                raise Exception(
                    "Site {site} has no {env} environment"
                    .format(site=site, env=selected_environment)
                )

            for setting, value in api.env.sites[site][selected_environment].iteritems():
                api.env[setting] = value

            api.env.site = site

        return func(*args, **kwargs)
    return wrapper



#!/usr/bin/env python
from setuptools import find_packages, setup
from fabliip import __version__

setup(
    name='fabliip',
    version=__version__,
    packages=find_packages(),
    description='Set of Fabric functions to help deploying websites.',
    long_description=open("README.rst", 'r').read().decode('utf-8'),
    author='Sylvain Fankhauser',
    author_email='sylvain.fankhauser@liip.ch',
    url='https://github.com/sephii/fabliip',
    install_requires=['fabric'],
)



from distutils.version import LooseVersion
import glob
import os


def get_version_files(from_version, to_version, directory):
    """
    Get a list of the files named after a version number (eg. 0.1.py, 0.1.2.py,
    etc) from the given directory and return a sorted list of tuples (version,
    file) of files which version number match the from_version (non inclusive)
    and to_version (inclusive) limits.

    Arguments:
    from_version -- A string representing the low version number (eg. 1.0)
    to_version -- A string representing the high version number (eg. 1.2.6)
    directory -- The path to the directory that holds the version files
    """

    files = []

    from_version = LooseVersion(from_version)
    to_version = LooseVersion(to_version)

    for file in glob.glob(os.path.join(directory, '*.py')):
        version, _ = os.path.splitext(os.path.basename(file))
        v = LooseVersion(version)

        if from_version < v <= to_version:
            files.append((v, file))

    files = sorted(files, key=lambda version: version[0])

    return files
